"""
_ExpressMerge_

Splitting algorithm for express merging
"""

import logging
import threading
import time

from WMCore.WMBS.File import File

from WMCore.JobSplitting.JobFactory import JobFactory
from WMCore.DAOFactory import DAOFactory
from WMCore.Services.UUIDLib import makeUUID


class ExpressMerge(JobFactory):
    """
    Split jobs by set of files

    """
    def algorithm(self, groupInstance = None, jobInstance = None,
                  *args, **kwargs):
        """
        _algorithm_

        A file based splitting algorithm

        """
        # extract some global scheduling parameters
        self.jobNamePrefix = kwargs.get('jobNamePrefix', "ExpressMerge")
        self.maxInputSize = kwargs['maxInputSize']
        self.maxInputFiles = kwargs['maxInputFiles']
        self.maxLatency = kwargs['maxLatency']
        self.currentTime = time.time()

        self.createdGroup = False

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package = "T0.WMBS",
                                logger = logging,
                                dbinterface = myThread.dbi)

        # data discovery
        getFilesDAO = daoFactory(classname = "Subscriptions.GetAvailableExpressMergeFiles")
        availableFiles = getFilesDAO.execute(self.subscription["id"])

        # nothing to do, stop immediately
        if len(availableFiles) == 0:
            return

        # sort by lumi
        filesByLumi = {}
        for result in availableFiles:
            lumi = result['lumi']
            if lumi in filesByLumi:
                filesByLumi[lumi].append(result)
            else:
                filesByLumi[lumi] = [ result ]

        self.defineJobs(filesByLumi)

        return


    def defineJobs(self, filesByLumi):
        """
        _defineJobs_

        schedule jobs

        """
        logging.debug("defineJobs(): Running...")

        #
        # actual scheduling algorith
        #
        # criteria:
        #     1. a lumi section should not sit around too long, waiting to get merged
        #     2. if possible, we want to merge in order of lumi sections (no holes)
        #     3. avoid too many small files (merge as many lumi sections as possible)
        #     4. don't merge too many lumi sections as the jobs run too long
        #     5. last, don't produce too big files
        #
        #     merge whatever we have (without holes) if oldest lumi section older than self.maxLatency
        #     if self.maxLatency is 0, merge lumi by lumi
        #

        lastLumi = 0
        jobSizeTotal = 0
        jobFileList = []

        for lumi in sorted(filesByLumi.keys()):

            lumiFileList = filesByLumi[lumi]

            # find oldest file in the lumi
            lumiDoneTime = 0
            for fileInfo in lumiFileList:
                if fileInfo['insert_time'] > lumiDoneTime:
                    lumiDoneTime = fileInfo['insert_time']
            lumiAge = self.currentTime - lumiDoneTime

            # calculate lumi size and new total size and file count
            lumiSizeTotal = 0
            for fileInfo in lumiFileList:
                lumiSizeTotal += fileInfo['filesize']

            newSizeTotal = jobSizeTotal + lumiSizeTotal
            newFileCount = len(jobFileList) + len(lumiFileList)

            # first lumi, if not old enough bail out
            if len(jobFileList) == 0:
                if lumiAge > self.maxLatency:
                    jobFileList.extend(lumiFileList)
                    lastLumi = lumi
                    jobSizeTotal = lumiSizeTotal
                else:
                    break
            # if self.maxLatency 0, just expressmerge lumi by lumi
            elif self.maxLatency == 0:
                self.createJob(jobFileList, jobSizeTotal)
                jobFileList = lumiFileList
                lastLumi = lumi
                jobSizeTotal = lumiSizeTotal
            # observe strict lumi order and sequence broken => expressmerge
            # triggers new age check on next out of sequence lumi
            # bail if not old enough
            elif lumi != lastLumi + 1:
                self.createJob(jobFileList, jobSizeTotal)
                if lumiAge > self.maxLatency:
                    jobFileList = lumiFileList
                    lastLumi = lumi
                    jobSizeTotal = lumiSizeTotal
                else:
                    jobFileList = []
                    break
            # if below limits just add to expressmerge job
            elif newFileCount <= self.maxInputFiles and \
                     newSizeTotal <= self.maxInputSize:
                jobFileList.extend(lumiFileList)
                lastLumi = lumi
                jobSizeTotal = newSizeTotal
            # over limits => expressmerge
            else:
                self.createJob(jobFileList, jobSizeTotal)
                jobFileList = lumiFileList
                lastLumi = lumi
                jobSizeTotal = lumiSizeTotal

        # sequential leftovers that are old enough
        if len(jobFileList) > 0:
            self.createJob(jobFileList, jobSizeTotal)

        return


    def createJob(self, fileList, jobSize):
        """
        _createJob_

        create an express merge job for
        the passed in list of files

        """
        if not self.createdGroup:
            self.newGroup()
            self.createdGroup = True

        self.newJob(name = "%s-%s" % (self.jobNamePrefix, makeUUID()))

        largestFile = 0
        for fileInfo in fileList:
            largestFile = max(largestFile, fileInfo['filesize'])
            f = File(id = fileInfo['id'],
                     lfn = fileInfo['lfn'])
            f.setLocation(fileInfo['location'], immediateSave = False)
            self.currentJob.addFile(f)

        # job time based on
        #   - 5 min initialization
        #   - 5MB/s merge speed
        #   - checksum calculation at 5MB/s
        #   - stageout at 5MB/s
        # job disk based on
        #  - input for largest file on local disk
        #  - output on local disk (factor 1)
        jobTime = 300 + (jobSize*3)/5000000
        self.currentJob.addResourceEstimates(jobTime = jobTime,
                                             disk = (jobSize+largestFile)/1024,
                                             memory = 1000)

        return



"""
_GetRunStreamDone_

Oracle implementation of GetRunStreamDone

Returns run/stream combinations that are fully processed

"""

from WMCore.Database.DBFormatter import DBFormatter

class GetRunStreamDone(DBFormatter):

    def execute(self, conn = None, transaction = False):

        returnList = []

        sql = """SELECT run_stream_done.run_id AS run,
                        stream.name AS stream
                 FROM run_stream_done
                 INNER JOIN stream ON
                   stream.id = run_stream_done.stream_id
                 LEFT OUTER JOIN run_stream_fileset_assoc ON
                   run_stream_fileset_assoc.run_id = run_stream_done.run_id AND
                   run_stream_fileset_assoc.stream_id = run_stream_done.stream_id
                 WHERE checkForZeroState(run_stream_done.in_datasvc) = 0
                 AND run_stream_fileset_assoc.run_id IS NULL
                 """

        results = self.dbi.processData(sql, binds = {}, conn = conn,
                                       transaction = transaction)[0].fetchall()

        for result in results:
            returnList.append( { 'run': result[0],
                                 'stream': result[1] } )

        sql = """SELECT run_stream_done.run_id AS run,
                        stream.name AS stream
                 FROM run_stream_done
                 INNER JOIN stream ON
                   stream.id = run_stream_done.stream_id
                 INNER JOIN run_stream_fileset_assoc ON
                   run_stream_fileset_assoc.run_id = run_stream_done.run_id AND
                   run_stream_fileset_assoc.stream_id = run_stream_done.stream_id
                 INNER JOIN wmbs_subscription ON
                   wmbs_subscription.fileset = run_stream_fileset_assoc.fileset
                 INNER JOIN wmbs_workflow ON
                   wmbs_workflow.id = wmbs_subscription.workflow
                 INNER JOIN wmbs_workflow runstream_workflow ON
                   runstream_workflow.name = wmbs_workflow.name
                 INNER JOIN wmbs_subscription runstream_subscription ON
                   runstream_subscription.workflow = runstream_workflow.id
                 WHERE checkForZeroState(run_stream_done.in_datasvc) = 0
                 GROUP BY run_stream_done.run_id,
                          stream.name
                 HAVING SUM(runstream_subscription.finished) = COUNT(*)
                 """

        results = self.dbi.processData(sql, binds = {}, conn = conn,
                                       transaction = transaction)[0].fetchall()

        for result in results:
            returnList.append( { 'run': result[0],
                                 'stream': result[1] } )

        return returnList



"""
_CheckClosedLumis_

Oracle implementation of RunLumiCloseout.CheckClosedLumis

Created on Nov 29, 2012

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter

class CheckClosedLumis(DBFormatter):
    """
    _CheckClosedLumis_

    Cross-checks the number of closed lumis for
    every stream in a run with the lumicount
    in the run table
    """

    sql = """SELECT stream_id,
                    COUNT(*) AS closed_lumi_count,
                    MAX(lumi_id) AS max_lumi,
                    (SELECT lumicount
                     FROM run
                     WHERE run_id = :RUN_ID) AS expected_lumi_count
             FROM lumi_section_closed
             WHERE run_id = :RUN_ID
             GROUP BY stream_id
          """

    def execute(self, run, conn = None, transaction = False):
        """
        _execute_

        Basic execute query plus formatDict
        """
        result = self.dbi.processData(self.sql, {'RUN_ID': run}, conn = conn,
                                      transaction = transaction)

        return self.formatDict(result)



"""
_MarkWorkflowsInjected_

Oracle implementation of MarkWorkflowsInjected

Implementation depends on whether we assume streamer
notifications to P5 or not.

with P5 streamer notifications:

 - Express sets a workflow to injected when the fileset
   is closed and all streamers are fed into the Tier0
 - Repack has the same requirements and in additions
   all datasets for that workflow/stream had their
   PromptReco released

without P5 streamer notifications:
 - Express is set to injected immediately
 - Repack needs all datasets for that workflow/stream
   PromptReco released before the workflow is set
   to injected

"""

from WMCore.Database.DBFormatter import DBFormatter

class MarkWorkflowsInjected(DBFormatter):

    sqlExpressNotify = \
       """UPDATE wmbs_workflow
          SET injected = 1
          WHERE name in (
            SELECT wmbs_workflow.name
            FROM run_stream_fileset_assoc
              INNER JOIN wmbs_subscription ON
                wmbs_subscription.fileset = run_stream_fileset_assoc.fileset
              INNER JOIN wmbs_fileset ON
                wmbs_fileset.id = wmbs_subscription.fileset AND
                wmbs_fileset.open = 0
              INNER JOIN wmbs_workflow ON
                wmbs_workflow.id = wmbs_subscription.workflow AND
                wmbs_workflow.injected = 0
              INNER JOIN run_stream_style_assoc ON
                run_stream_style_assoc.run_id = run_stream_fileset_assoc.run_id AND
                run_stream_style_assoc.stream_id =   run_stream_fileset_assoc.stream_id AND
                run_stream_style_assoc.style_id = (SELECT id FROM processing_style WHERE name = 'Express')
              LEFT OUTER JOIN streamer ON
                streamer.run_id = run_stream_fileset_assoc.run_id AND
                streamer.stream_id = run_stream_fileset_assoc.stream_id AND
                checkForZeroState(streamer.deleted) = 0
            WHERE streamer.run_id IS NULL
            GROUP BY wmbs_workflow.name
          )
          """

    sqlRepackNotify = \
       """UPDATE wmbs_workflow
          SET injected = 1
          WHERE name in (
            SELECT wmbs_workflow.name
            FROM run_stream_fileset_assoc
              INNER JOIN wmbs_subscription ON
                wmbs_subscription.fileset = run_stream_fileset_assoc.fileset
              INNER JOIN wmbs_fileset ON
                wmbs_fileset.id = wmbs_subscription.fileset AND
                wmbs_fileset.open = 0
              INNER JOIN wmbs_workflow ON
                wmbs_workflow.id = wmbs_subscription.workflow AND
                wmbs_workflow.injected = 0
              INNER JOIN run_stream_style_assoc ON
                run_stream_style_assoc.run_id = run_stream_fileset_assoc.run_id AND
                run_stream_style_assoc.stream_id =   run_stream_fileset_assoc.stream_id AND
                run_stream_style_assoc.style_id = (SELECT id FROM processing_style WHERE name = 'Bulk')
              LEFT OUTER JOIN streamer ON
                streamer.run_id = run_stream_fileset_assoc.run_id AND
                streamer.stream_id = run_stream_fileset_assoc.stream_id AND
                checkForZeroState(streamer.deleted) = 0
              INNER JOIN run_primds_stream_assoc ON
                run_primds_stream_assoc.run_id = run_stream_fileset_assoc.run_id AND
                run_primds_stream_assoc.stream_id =   run_stream_fileset_assoc.stream_id
              LEFT OUTER JOIN reco_config ON
                reco_config.run_id = run_stream_fileset_assoc.run_id AND
                reco_config.primds_id = run_primds_stream_assoc.primds_id
            WHERE streamer.run_id IS NULL
            GROUP BY wmbs_workflow.name
            HAVING COUNT(reco_config.run_id) = COUNT(*)
          )
          """

    sqlExpressNoNotify = \
       """UPDATE wmbs_workflow
          SET injected = 1
          WHERE name in (
            SELECT wmbs_workflow.name
            FROM run_stream_fileset_assoc
              INNER JOIN wmbs_subscription ON
                wmbs_subscription.fileset = run_stream_fileset_assoc.fileset
              INNER JOIN wmbs_workflow ON
                wmbs_workflow.id = wmbs_subscription.workflow AND
                wmbs_workflow.injected = 0
              INNER JOIN run_stream_style_assoc ON
                run_stream_style_assoc.run_id = run_stream_fileset_assoc.run_id AND
                run_stream_style_assoc.stream_id =   run_stream_fileset_assoc.stream_id AND
                run_stream_style_assoc.style_id = (SELECT id FROM processing_style WHERE name = 'Express')
            GROUP BY wmbs_workflow.name
          )
          """

    sqlRepackNoNotify = \
       """UPDATE wmbs_workflow
          SET injected = 1
          WHERE name in (
            SELECT wmbs_workflow.name
            FROM run_stream_fileset_assoc
              INNER JOIN wmbs_subscription ON
                wmbs_subscription.fileset = run_stream_fileset_assoc.fileset
              INNER JOIN wmbs_workflow ON
                wmbs_workflow.id = wmbs_subscription.workflow AND
                wmbs_workflow.injected = 0
              INNER JOIN run_stream_style_assoc ON
                run_stream_style_assoc.run_id = run_stream_fileset_assoc.run_id AND
                run_stream_style_assoc.stream_id =   run_stream_fileset_assoc.stream_id AND
                run_stream_style_assoc.style_id = (SELECT id FROM processing_style WHERE name = 'Bulk')
              INNER JOIN run_primds_stream_assoc ON
                run_primds_stream_assoc.run_id = run_stream_fileset_assoc.run_id AND
                run_primds_stream_assoc.stream_id =   run_stream_fileset_assoc.stream_id
              LEFT OUTER JOIN reco_config ON
                reco_config.run_id = run_stream_fileset_assoc.run_id AND
                reco_config.primds_id = run_primds_stream_assoc.primds_id
            GROUP BY wmbs_workflow.name
            HAVING COUNT(reco_config.run_id) = COUNT(*)
          )
          """

    def execute(self, streamerNotification, conn = None, transaction = False):

        if streamerNotification:

            self.dbi.processData(self.sqlExpressNotify, binds = {}, conn = conn,
                                 transaction = transaction)

            self.dbi.processData(self.sqlRepackNotify, binds = {}, conn = conn,
                                 transaction = transaction)

        else:

            self.dbi.processData(self.sqlExpressNoNotify, binds = {}, conn = conn,
                                 transaction = transaction)

            self.dbi.processData(self.sqlRepackNoNotify, binds = {}, conn = conn,
                                 transaction = transaction)

        return



"""
_HaveJobGroup_

Oracle implementation of HaveJobGroup

For a given subscription check if there is an existing job group
"""

from WMCore.Database.DBFormatter import DBFormatter

class HaveJobGroup(DBFormatter):

    sql = """SELECT 1
             FROM wmbs_jobgroup
             WHERE wmbs_jobgroup.subscription = :subscription
             AND ROWNUM = 1
             """

    def execute(self, subscription, conn = None, transaction = False):

        results = self.dbi.processData(self.sql, { 'subscription' : subscription },
                                       conn = conn, transaction = transaction)[0].fetchall()

        return ( len(results) > 0 and results[0][0] == 1 )



"""
_MarkTrackedWorkflowMonitoring_

Oracle implementation of MarkTrackedWorkflowMonitoring.

Sets a workflow as tracked=1 in workflow_monitoring table,
should be used after successful upload to Couch.

"""

from WMCore.Database.DBFormatter import DBFormatter

class MarkTrackedWorkflowMonitoring(DBFormatter):

    def execute(self, workflowId, conn = None, transaction = False):

        sql = """UPDATE workflow_monitoring
                 SET tracked = 1
                 WHERE workflow = :WORKFLOW_ID"""

        binds = { 'WORKFLOW_ID' : workflowId }

        self.dbi.processData(sql, binds, conn = conn,
                             transaction = transaction)

        return



"""
_InsertRun_

Oracle implementation of InsertRun

"""

from WMCore.Database.DBFormatter import DBFormatter

class InsertRun(DBFormatter):

    def execute(self, binds, conn = None, transaction = False):

        sql = """INSERT INTO run
                 (RUN_ID, HLTKEY)
                 SELECT :RUN,
                        :HLTKEY
                 FROM DUAL
                 WHERE NOT EXISTS (
                   SELECT * FROM run WHERE run_id = :RUN
                 )"""

        self.dbi.processData(sql, binds, conn = conn,
                             transaction = transaction)

        return



#!/usr/bin/env python
"""
_Express_t_

Express job splitting test

"""

import unittest
import threading
import logging
import time

from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMCore.DataStructs.Run import Run

from WMCore.DAOFactory import DAOFactory
from WMCore.JobSplitting.SplitterFactory import SplitterFactory
from WMCore.Services.UUIDLib import makeUUID
from WMQuality.TestInit import TestInit


class ExpressTest(unittest.TestCase):
    """
    _ExpressTest_

    Test for Express job splitter
    """

    def setUp(self):
        """
        _setUp_

        """
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()

        self.testInit.setSchema(customModules = ["WMComponent.DBS3Buffer", "T0.WMBS"])

        self.splitterFactory = SplitterFactory(package = "T0.JobSplitting")

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package = "T0.WMBS",
                                logger = logging,
                                dbinterface = myThread.dbi)

        myThread.dbi.processData("""INSERT INTO wmbs_location
                                    (id, site_name, state, state_time)
                                    VALUES (1, 'SomeSite', 1, 1)
                                    """, transaction = False)
        myThread.dbi.processData("""INSERT INTO wmbs_pnns
                                    (id, pnn)
                                    VALUES (2, 'SomePNN')
                                    """, transaction = False)

        myThread.dbi.processData("""INSERT INTO wmbs_location_pnns
                                    (location, pnn)
                                    VALUES (1, 2)
                                    """, transaction = False)


        insertRunDAO = daoFactory(classname = "RunConfig.InsertRun")
        insertRunDAO.execute(binds = { 'RUN' : 1,
                                       'HLTKEY' : "someHLTKey" },
                             transaction = False)

        insertLumiDAO = daoFactory(classname = "RunConfig.InsertLumiSection")
        for lumi in [1, 2]:
            insertLumiDAO.execute(binds = { 'RUN' : 1,
                                            'LUMI' : lumi },
                                  transaction = False)

        insertStreamDAO = daoFactory(classname = "RunConfig.InsertStream")
        insertStreamDAO.execute(binds = { 'STREAM' : "Express" },
                                transaction = False)

        insertStreamFilesetDAO = daoFactory(classname = "RunConfig.InsertStreamFileset")
        insertStreamFilesetDAO.execute(1, "Express", "TestFileset1")

        self.fileset1 = Fileset(name = "TestFileset1")
        self.fileset1.load()

        workflow1 = Workflow(spec = "spec.xml", owner = "hufnagel", name = "TestWorkflow1", task="Test")
        workflow1.create()

        self.subscription1  = Subscription(fileset = self.fileset1,
                                           workflow = workflow1,
                                           split_algo = "Express",
                                           type = "Express")
        self.subscription1.create()

        # keep for later
        self.insertClosedLumiDAO = daoFactory(classname = "RunLumiCloseout.InsertClosedLumi")
        self.releaseExpressDAO = daoFactory(classname = "Tier0Feeder.ReleaseExpress")
        self.currentTime = int(time.time())

        return

    def tearDown(self):
        """
        _tearDown_

        """
        self.testInit.clearDatabase()

        return

    def finalCloseLumis(self):
        """
        _finalCloseLumis_

        """
        myThread = threading.currentThread()

        myThread.dbi.processData("""UPDATE lumi_section_closed
                                    SET close_time = 1
                                    """, transaction = False)

        return

    def getNumActiveSplitLumis(self):
        """
        _getNumActiveSplitLumis_

        helper function that counts the number of active split lumis
        """
        myThread = threading.currentThread()

        results = myThread.dbi.processData("""SELECT COUNT(*)
                                              FROM lumi_section_split_active
                                              """, transaction = False)[0].fetchall()

        return results[0][0]

    def test00(self):
        """
        _test00_

        Test that the job name prefix feature works
        Test event threshold (single job creation)

        Test that only closed lumis are used

        Test check on express release

        """
        insertClosedLumiBinds = []
        for lumi in [1]:
            filecount = 2
            for i in range(filecount):
                newFile = File(makeUUID(), size = 1000, events = 100)
                newFile.addRun(Run(1, *[lumi]))
                newFile.setLocation("SomePNN", immediateSave = False)
                newFile.create()
                self.fileset1.addFile(newFile)
                insertClosedLumiBinds.append( { 'RUN' : 1,
                                                'LUMI' : lumi,
                                                'STREAM' : "Express",
                                                'FILECOUNT' : filecount,
                                                'INSERT_TIME' : self.currentTime,
                                                'CLOSE_TIME' : 0 } )
        self.fileset1.commit()

        jobFactory = self.splitterFactory(package = "WMCore.WMBS",
                                          subscription = self.subscription1)

        jobGroups = jobFactory(maxInputEvents = 200, maxInputRate = 23000)

        self.assertEqual(len(jobGroups), 0,
                         "ERROR: JobFactory should have returned no JobGroup")

        self.insertClosedLumiDAO.execute(binds = insertClosedLumiBinds,
                                         transaction = False)

        jobGroups = jobFactory(maxInputEvents = 200, maxInputRate = 23000)

        self.assertEqual(len(jobGroups), 0,
                         "ERROR: JobFactory should have returned no JobGroup")

        self.finalCloseLumis()

        jobGroups = jobFactory(maxInputEvents = 200, maxInputRate = 23000)

        self.assertEqual(len(jobGroups), 0,
                         "ERROR: JobFactory should have returned no JobGroup")

        self.releaseExpressDAO.execute(binds = { 'RUN' : 1 }, transaction = False)

        jobGroups = jobFactory(maxInputEvents = 200, maxInputRate = 23000)

        self.assertEqual(len(jobGroups), 1,
                         "ERROR: JobFactory didn't return one JobGroup")

        self.assertEqual(len(jobGroups[0].jobs), 1,
                         "ERROR: JobFactory didn't create a single job")

        job = jobGroups[0].jobs[0]
        self.assertTrue(job['name'].startswith("Express-"),
                        "ERROR: Job has wrong name")

        self.assertEqual(self.getNumActiveSplitLumis(), 0,
                         "ERROR: Split lumis were created")

        return

    def test01(self):
        """
        _test01_

        Test event threshold (multiple job creation)

        """
        insertClosedLumiBinds = []
        for lumi in [1]:
            filecount = 2
            for i in range(filecount):
                newFile = File(makeUUID(), size = 1000, events = 100)
                newFile.addRun(Run(1, *[lumi]))
                newFile.setLocation("SomePNN", immediateSave = False)
                newFile.create()
                self.fileset1.addFile(newFile)
                insertClosedLumiBinds.append( { 'RUN' : 1,
                                                'LUMI' : lumi,
                                                'STREAM' : "Express",
                                                'FILECOUNT' : filecount,
                                                'INSERT_TIME' : self.currentTime,
                                                'CLOSE_TIME' : self.currentTime } )
        self.fileset1.commit()

        jobFactory = self.splitterFactory(package = "WMCore.WMBS",
                                          subscription = self.subscription1)

        self.insertClosedLumiDAO.execute(binds = insertClosedLumiBinds,
                                         transaction = False)

        self.releaseExpressDAO.execute(binds = { 'RUN' : 1 }, transaction = False)

        jobGroups = jobFactory(maxInputEvents = 199, maxInputRate = 23000)

        self.assertEqual(len(jobGroups[0].jobs), 2,
                         "ERROR: JobFactory didn't create two jobs")

        self.assertEqual(self.getNumActiveSplitLumis(), 1,
                         "ERROR: Didn't create a single split lumi")

        return

    def test02(self):
        """
        _test02_

        Test multi lumis

        """
        insertClosedLumiBinds = []
        for lumi in [1, 2]:
            filecount = 1
            for i in range(filecount):
                newFile = File(makeUUID(), size = 1000, events = 100)
                newFile.addRun(Run(1, *[lumi]))
                newFile.setLocation("SomePNN", immediateSave = False)
                newFile.create()
                self.fileset1.addFile(newFile)
                insertClosedLumiBinds.append( { 'RUN' : 1,
                                                'LUMI' : lumi,
                                                'STREAM' : "Express",
                                                'FILECOUNT' : filecount,
                                                'INSERT_TIME' : self.currentTime,
                                                'CLOSE_TIME' : self.currentTime } )
        self.fileset1.commit()

        jobFactory = self.splitterFactory(package = "WMCore.WMBS",
                                          subscription = self.subscription1)

        self.insertClosedLumiDAO.execute(binds = insertClosedLumiBinds,
                                         transaction = False)

        self.releaseExpressDAO.execute(binds = { 'RUN' : 1 }, transaction = False)

        jobGroups = jobFactory(maxInputEvents = 100, maxInputRate = 23000)

        self.assertEqual(len(jobGroups[0].jobs), 2,
                         "ERROR: JobFactory didn't create two jobs")

        self.assertEqual(self.getNumActiveSplitLumis(), 0,
                         "ERROR: Split lumis were created")

        return

if __name__ == '__main__':
    unittest.main()



"""
_UpdateOfflineFileStatus_

Oracle implementation of UpdateOfflineFileStatus

"""

from WMCore.Database.DBFormatter import DBFormatter

class UpdateOfflineFileStatus(DBFormatter):

    def execute(self, binds, conn = None, transaction = False):

        sql = """UPDATE file_transfer_status_offline
                 SET t0_repacked_time = CURRENT_TIMESTAMP,
                     repacked_retrieve = 1
                 WHERE p5_fileid = :P5_ID
                 AND t0_repacked_time IS NULL
                 AND repacked_retrieve is NULL
                 """

        self.dbi.processData(sql, binds, conn = conn,
                             transaction = transaction)

        return



"""
_InsertExpressConfig_

Oracle implementation of InsertExpressConfig

"""

from WMCore.Database.DBFormatter import DBFormatter

class InsertExpressConfig(DBFormatter):

    def execute(self, binds, conn = None, transaction = False):

        sql = """INSERT INTO express_config
                 (RUN_ID, STREAM_ID, PROC_VERSION, WRITE_TIERS, WRITE_DQM, GLOBAL_TAG,
                  MAX_RATE, MAX_EVENTS, MAX_SIZE, MAX_FILES, MAX_LATENCY, DQM_INTERVAL,
                  CMSSW_ID, SCRAM_ARCH, RECO_CMSSW_ID,RECO_SCRAM_ARCH, DATA_TYPE,
                  MULTICORE, ALCA_SKIM, DQM_SEQ)
                 VALUES (:RUN,
                         (SELECT id FROM stream WHERE name = :STREAM),
                         :PROC_VER,
                         :WRITE_TIERS,
                         :WRITE_DQM,
                         :GLOBAL_TAG,
                         :MAX_RATE,
                         :MAX_EVENTS,
                         :MAX_SIZE,
                         :MAX_FILES,
                         :MAX_LATENCY,
                         :DQM_INTERVAL,
                         (SELECT id FROM cmssw_version WHERE name = :CMSSW),
                         :SCRAM_ARCH,
                         (SELECT id FROM cmssw_version WHERE name = :RECO_CMSSW),
                         :RECO_SCRAM_ARCH,
                         :DATA_TYPE,
                         :MULTICORE,
                         :ALCA_SKIM,
                         :DQM_SEQ)
                 """

        self.dbi.processData(sql, binds, conn = conn,
                             transaction = transaction)

        return



"""
_InsertWorkflowMonitoring_

Oracle implementation of InsertWorkflowMonitoring

Important bit is that the only unique identifier per workflow
is the top level fileset, which is the input to the first task.
Supports as input one or more fileset.ids as a list.

"""

from WMCore.Database.DBFormatter import DBFormatter

class InsertWorkflowMonitoring(DBFormatter):

    def execute(self, filesets, conn = None, transaction = False):

        sql = """INSERT INTO workflow_monitoring (WORKFLOW) 
                 VALUES ((SELECT workflow
                          FROM wmbs_subscription
                          WHERE fileset = :FILESET))
                 """

        binds = []
        for fileset in filesets: 
            binds.append( { 'FILESET' : fileset } )

        self.dbi.processData(sql, binds, conn = conn,
                             transaction = transaction)

        return



"""
_InsertDatasetTrigger_

Oracle implementation of InsertDatasetTrigger

"""

from WMCore.Database.DBFormatter import DBFormatter

class InsertDatasetTrigger(DBFormatter):

    def execute(self, binds, conn = None, transaction = False):

        sql = """INSERT INTO run_trig_primds_assoc
                 (RUN_ID, TRIG_ID, PRIMDS_ID)
                 VALUES (:RUN,
                         (SELECT id FROM trigger_label WHERE name = :TRIG),
                         (SELECT id FROM primary_dataset WHERE name = :PRIMDS))
                 """

        self.dbi.processData(sql, binds, conn = conn,
                             transaction = transaction)

        return



"""
_InsertPhEDExConfig_

Oracle implementation of InsertPhEDExConfig

"""

from WMCore.Database.DBFormatter import DBFormatter

class InsertPhEDExConfig(DBFormatter):

    sql = """INSERT INTO phedex_config
             (RUN_ID, PRIMDS_ID, ARCHIVAL_NODE_ID, TAPE_NODE_ID, DISK_NODE_ID, DISK_NODE_RECO_ID)
             VALUES (:RUN,
                    (SELECT id FROM primary_dataset WHERE name = :PRIMDS),
                    (SELECT id FROM storage_node WHERE name = :ARCHIVAL_NODE),
                    (SELECT id FROM storage_node WHERE name = :TAPE_NODE),
                    (SELECT id FROM storage_node WHERE name = :DISK_NODE),
                    (SELECT id FROM storage_node WHERE name = :DISK_NODE_RECO))
             """

    def execute(self, binds, conn = None, transaction = False):

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)

        return



"""
_CheckActiveSplitLumis_

Oracle implementation of CheckActiveSplitLumis

Check active split lumis for completetion and delete if complete.

An active split lumi is complete if the number of complete
file records is equal to the total numbers of streamer files
in that lumi.

"""
import time

from WMCore.Database.DBFormatter import DBFormatter

class CheckActiveSplitLumis(DBFormatter):

    def execute(self, conn = None, transaction = False):

        sql = """DELETE FROM lumi_section_split_active
                 WHERE ( lumi_section_split_active.subscription,
                         lumi_section_split_active.run_id,
                         lumi_section_split_active.lumi_id )
                 IN (
                   SELECT lumi_section_split_active.subscription,
                          lumi_section_split_active.run_id,
                          lumi_section_split_active.lumi_id
                   FROM lumi_section_split_active
                   INNER JOIN wmbs_sub_files_complete ON
                     wmbs_sub_files_complete.subscription = lumi_section_split_active.subscription
                   INNER JOIN wmbs_file_runlumi_map ON
                     wmbs_file_runlumi_map.fileid = wmbs_sub_files_complete.fileid AND
                     wmbs_file_runlumi_map.run = lumi_section_split_active.run_id AND
                     wmbs_file_runlumi_map.lumi = lumi_section_split_active.lumi_id
                   GROUP BY lumi_section_split_active.subscription,
                            lumi_section_split_active.run_id,
                            lumi_section_split_active.lumi_id
                   HAVING COUNT(*) = MAX(lumi_section_split_active.nfiles)
                 )
                 """

        self.dbi.processData(sql, binds = {}, conn = conn,
                             transaction = transaction)

        return
